/* Copyright (c) 2005, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include <unistd.h>

#include <curses.h>

#include "mazegen.h"

char *alg_name[] = {
  "Backtracer",
  "Growtree",
  0};

MAZE *(*alg_func[])(int x, int y, vec start, vec end, int step) = {
  mazegen_backtracer,
    mazegen_growtree,
    0};

char *numsnames[] = { "X", "Y", "alg", "step", "trace", 0};

int alg_count;
    

// Screen atributes, for display
int wall;
int unreachable;
int reachable;
int headercol;
int dood;
int startend;
int text;
int solution;

int main() {
  
  // Init our random numbers
  srandom(time(0));

  initscr();
  
  start_color();
  noecho();
  cbreak();
  nonl();
  keypad(stdscr, TRUE);
  curs_set(1);
  
  //colors
  init_pair(1, COLOR_WHITE, COLOR_BLACK);
  init_pair(2, COLOR_RED, COLOR_BLACK);
  init_pair(3, COLOR_BLUE, COLOR_BLACK);
  init_pair(4, COLOR_WHITE, COLOR_BLUE);
  init_pair(5, COLOR_GREEN, COLOR_BLACK);
  init_pair(6, COLOR_WHITE, COLOR_BLACK);
  init_pair(7, COLOR_RED, COLOR_BLACK);
  wall = COLOR_PAIR(1) | A_BOLD;
  unreachable = COLOR_PAIR(2) | A_BOLD;
  reachable =  COLOR_PAIR(3) | A_DIM;
  headercol = COLOR_PAIR(4) | A_BOLD;
  dood = COLOR_PAIR(7);
  startend = COLOR_PAIR(5) | A_BOLD;
  text = COLOR_PAIR(6);
  solution = COLOR_PAIR(5);

  vec res;
  res.x = res.y = 0;
  int alg = 0;
  // 0 = backtracer, 1 = growtree
  int step = 0;
  // Step by step?
  int trace = 0;

  for (alg_count = 0; alg_name[alg_count] != NULL; alg_count++);

  int userinput = -1;
  // -1 = none, 0 == x, 1 == y, 2 == alg, 3 == step, 4 == trace

  
  MAZE *world = NULL;
  int redraw = 1;

  int sx=0, sy=0,x=0,y=0;
  int mazex=0, mazey=0;
  vec you;
  while (1) {
    if (alg < 0) {
      alg = 0;
    } else if (alg >= alg_count) {
      alg = alg_count - 1;
    }
    if (step < 0) {
      step = 0;
    } else if (step > 1) {
      step = 1;
    }

    if (res.x != 0 || res.y != 0) {
      x = res.x;
      y = res.y;
    } else {
      getmaxyx(stdscr,sy,sx);
      if (sx <= 0 || sy <= 0) {
	x = 38; // 80x25 default if i can't get the resolution
	y = 10;
      } else {
	x = (sx / 2) - 1;
	y = (sy / 2) - 1;
      }
    }
    if (redraw) {
      vec start, end;
      start.x = 1;
      start.y = 0;
      end.x = x - 2;
      end.y = y-1;
      
      if (world != NULL)
	free(world);
      world = (*alg_func[alg])(x,y,start,end, step);
      you.x = world->enter.x; you.y = world->enter.y;
      mazex = x; mazey = y; // Write down the maze's dimensions, so when you start changing them while disping the old one, nothign breaks.
      redraw = 0;
    }
    //MAZE *world = mazegen_backtracer(x,y,start,end,0);
    //MAZE *world = newmaze(x,y);
    char *buf;
    buf = malloc(sx);
    snprintf(buf,sx, "Maze generation complete. (%ix%i). (%ix%i%s) (%s) (Step by step %s) (%s) %s", mazex,mazey, x,y,(res.x||res.y?"c":""), alg_name[alg], (step ? "on" : "off"), (trace?"Tracing":"Not tracing"), (userinput == -1 ? "" : numsnames[userinput]));
    dispmaze(world, you.x,you.y, buf);
    free(buf); // text displayed in dispmaze
    
    int input = getch();
    if (input == '\n' || input == 10 || input == 13) {
      userinput = -1;
      redraw = 1;
    } else if (input == 'q' || input == 'Q') {
      endwin();
      exit(0);
    } else if (input == 'x') {
      userinput = 0;
    } else if (input == 'X') {
      x = (sx / 2) - 1;
      res.x = 0;
    } else if (input == 'Y') {
      y = (sy / 2) - 1;
      res.y = 0;
    } else if (input == 'y') {
      userinput = 1;
    } else if (input == 'a') {
      userinput = 2;
    } else if (input == 's') {
      userinput = 3;
    } else if (input == 't') {
      userinput = 4;
    } else if (input == 8 || input == 127) { //backspace
      if (userinput == 0) {
	res.x /= 10;
      } else if (userinput == 1) {
	res.y /= 10;
      } else if (userinput == 2) {
	alg = 0;
      } else if (userinput == 3) {
	step = 0;
      } else if (userinput == 4) {
	trace = 0;
      }
    } else if (input >= '0' && input <= '9') {
      int num = input - '0';
      
      if (userinput == 0) {
	res.x = res.x*10 + num;
      } else if (userinput == 1) {
	res.y = res.y*10 + num;
      } else if (userinput == 2) {
	alg = num;
      } else if (userinput == 3) {
	step = num;
      } else if (userinput == 4) {
	trace = num;
      }
      // -1 = none, 0 == x, 1 == y, 2 == alg, 3 == step
    } else if (input == KEY_LEFT) {
      if (you.x > 0 && (world->path[you.y][you.x - 1] & E)) {
	you.x--;
	if (trace)
	  mset(world,you.x,you.y, SOL);
      }
    } else if (input == KEY_RIGHT) {
      if (you.x < world->x - 1 && (world->path[you.y][you.x] & E)) {
	you.x++;
	if (trace)
	  mset(world,you.x,you.y, SOL);
      }
    } else if (input == KEY_UP) {
      if (you.y > 0 && (world->path[you.y - 1][you.x] & S)) {
      	you.y--;
	if (trace)
	  mset(world,you.x,you.y, SOL);
      }
    } else if (input == KEY_DOWN) {
      if (you.y < world->y - 1 && (world->path[you.y][you.x] & S)) {
     	you.y++;
	if (trace)
	  mset(world,you.x,you.y, SOL);
      }
    }
    
  }
}

MAZE *newmaze(int x, int y) {
  MAZE *m;
  m = malloc(sizeof(MAZE));
  m->enter.x = m->enter.y = 0;
  m->exit.x = m->exit.y = 0;
  m->x = x;
  m->y = y;

  m->path = malloc( (1+y) * sizeof(unsigned char *) );
  int i;
  for (i = 0; i < y; i++) {
    m->path[i] = malloc( sizeof(char) * (x+1) );
    int i2;
    for (i2 = 0; i2 < x; i2++) {
      m->path[i][i2] = 0;
    }
  }
    

    
  return m;  
}

void dispmaze(MAZE *m, int fx, int fy, char *header) {
  int cx, cy;

  erase();
  
  // Draw a # border around the top/left edge of the border
  // Bottom and right get drawn in with individual cells
  attrset(wall);
  for (cx = 0; cx < m->x; cx++) {
    mvaddstr(1, cx * 2, "##");
  }
  mvaddstr(1, m->x * 2, "#");
  
  for (cy = 0; cy < m->y; cy++) {
    mvaddch(1 + cy * 2, 0, '#');
    mvaddch(2 + cy * 2, 0, '#');
  }
  mvaddch(1 + m->y * 2, 0, '#');


  for (cy = 0; cy < m->y; cy++) {
    for (cx = 0; cx < m->x; cx++) { 
      attrset(wall);
      mvaddch(2 + (2*cy), 2+ (2*cx),'#');
      mvaddstr(3 + (2*cy),1+ (2*cx),"##");
      int v = m->path[cy][cx];
      if (v&S) {
	attrset(text);
	mvaddch(3 + (cy*2), 1 + (cx*2), ' ');
      }
      if (v&E) {
	attrset(text);
	mvaddch(2 + (cy*2), 2 + (cx*2), ' ');
      }
      if (v&R) {
	if (v&SOL) {
	  attrset(solution);
	  mvaddch(2 + (cy*2), 1 + (cx*2), '*');
	} else {
	  attrset(reachable);
	  mvaddch(2 + (cy*2), 1 + (cx*2), '.');
	}
      } else {
	attrset(wall);
	mvaddch(2 + (cy*2), 1 + (cx*2), '#');
	//attrset(unreachable);
	//mvaddch(2 + (cy*2), 1 + (cx*2), 'x');
      }

    }
  }
  
  // Carve out the entrance/exit
  attrset(startend);
  mvaddch(2 + (m->enter.y*2), 1 + (m->enter.x*2), '<');
  mvaddch(2 + (m->exit.y*2), 1 + (m->exit.x*2), '>');
  
  // If fx and fy (focus x and y) aren't negative, show your d00d in the maze
  if (fx >= 0 && fy >= 0) {
    attrset(dood);
    mvaddch(2 + (fy*2), 1 + (fx*2), '@');
  }

  attrset(headercol);
  move(0,0);
  clrtoeol();
  addstr(header);
  refresh();
}

vec dirtovec(int dir) {
  vec v;
  v.x = 0;
  v.y = 0;
  switch (dir) {
  case N:
    v.y = -1;
    break;
  case E:
    v.x = 1;
    break;
  case S:
    v.y = 1;
    break;
  case W:
    v.x = -1;
    break;
  }
  return v;
}

void mset(MAZE *m, int x, int y, int val) {
  m->path[y][x] |= val;
}
void munset(MAZE *m, int x, int y, int val) {
  m->path[y][x] &= ~val;
}

void dig(MAZE *m, int x, int y, int dir) {
  switch (dir) {
  case N:
    mset(m,x,y - 1, R|S);
    break;
  case E:
    mset(m,x+1,y, R);
    mset(m,x,y, E);
    break;
  case S:
    mset(m,x,y+1,R);
    mset(m,x,y,S);
    break;
  case W:
    mset(m,x-1,y,R|E);
    break;
  default:
    break;
  }
}

char edge(MAZE *m, int x, int y, int dir) {
  vec d = dirtovec(dir);
  if ((x + d.x < 0) || (x + d.x >= m->x))
    return 1;
  if ((y + d.y < 0) || (y + d.y >= m->y))
    return 1;
  return 0;
}

// Returns a 6 char long string
// First char is the length of the next 5
// Next 5 are the #defined values of N,E,S,W, if they are valid moves
// (eg, not already reachable, and not edge of the world)
// You can feed (findvaliddirs+1)[ rand()%findvaliddirs[0] ] to get a random
//   valid direction.
// r=TRUE means reachables are okay, too
char *findvaliddirs(MAZE *m, int cx, int cy, int r)
{
  char dirs[5];
  char p = 0;
  if (!edge(m,cx,cy,N)) {
    if (!( m->path[cy-1][cx] & (R|START) ) | r) {
      dirs[(int)p] = N;
      p++;
    }
  }
  if (!edge(m,cx,cy,E)) {
    if (!( m->path[cy][cx+1] & (R|START) ) | r) {
      dirs[(int)p] = E;
      p++;
    }
  }
  if (!edge(m,cx,cy,S)) {
    if (!( m->path[cy+1][cx] & (R|START) ) | r) {
      dirs[(int)p] = S;
      p++;
    }
  }
  if (!edge(m,cx,cy,W)) {
    if (!( m->path[cy][cx-1] & (R|START) ) | r) {
      dirs[(int)p] = W;
      p++;
    }
  }
  char *export = malloc(sizeof(char)*6);
  export[0] = p;
  memcpy(export+1,&dirs,5);
  return export;
}

// Some functions to work with linked lists
LLIST *newllist() {
  LLIST *l = malloc(sizeof(LLIST));
  l->length = 0;
  l->cell = NULL;
  return l;
}

void destroyllist(LLIST *l) {
  if (l == NULL)
    return;

  struct CELL *now = l->cell;
  while (l->cell) {
    while (now->next != NULL)
      now = now->next;
    free(now);
  }
  free(l);
  return;
}

void addlink(LLIST *l, int pos, int x, int y, int value) {
  if (l == NULL)
    return;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return;

  struct CELL *add = malloc(sizeof(struct CELL));
  add->x = x; add->y = y; add->value = value; add->next = NULL;

  if (pos == 0) {
    struct CELL *orig = l->cell;
    l->cell = add;
    add->next = orig;
    l->length++; // Keep this count accurate.
    return;
  }

  int p;
  struct CELL *now = l->cell;
  
  for (p = 1; p < pos && now->next != NULL; p++)
    now = now->next;
  
  struct CELL *n; // Store the cell's current next d00d
  n = now->next;
  now->next = add; // Now, the next d00d of the cell is our addition
  add->next = n; // and, the next d00d of our addition is the regular cell.  addition to the middle of a list complete!
  l->length++; // Keep this count accurate.
  
}  
    
void remlink(LLIST *l, int pos) {
  if (l == NULL)
    return;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return;

  if (pos == 0) { // Removing the first in the list? Need to change the original struct, too, then.
    struct CELL *x = l->cell;
    l->cell = x->next;
    free(x);
    l->length--; // Keep this count accurate.
    return;
  }

  int i;
  struct CELL *now = l->cell;

  // Used in case we're deleting cell 1, in which case the 'last' cell (cell before the one being deleted) will point to element 0
  struct CELL *last;
  now = last = l->cell;

  for (i = 0; i < pos && now->next != NULL; i++) {
    last = now;
    now = now->next;
  }
  
  if (i == pos) {
    struct CELL *next = now->next;
    last->next = next; // The current (to be removed) cell is dropped from the linked list
    free(now); // And then dropped from system memory
    l->length--; // Keep this count accurate.
  }
  
}

struct CELL *getlink(LLIST *l, int pos) {
  if (l == NULL)
    return NULL;
  if (pos == -1 && l->length == 0)
    pos = 0;
  if (pos < 0)
    pos = l->length + pos + 1;
  if (pos < 0)
    return NULL;

  if (pos == 0) {
    return l->cell;
  }

  int p;
  struct CELL *now = l->cell;

  for (p = 0; p < pos && now->next != NULL; p++)
    now = now->next;

  return now;
}


void printllist(LLIST *l) {
  if (l == NULL)
    return;

  struct CELL *c;
  int i;
  for (i = 0; i < l->length; i++) {
    c = getlink(l,i);
    printf("pos% 3i:  x=%i, y=%i, val=%i\n", i,c->x,c->y,c->value);
  }
  printf("list length: %i\n",l->length);
  return;
  //////    
  struct CELL *now = l->cell;
  int p = 0;
  while (now != NULL) {
    printf("pos% 3i:  x=%i, y=%i, val=%i\n", p,now->x,now->y,now->value);
    p++;
    now = now->next;
  }
  printf("list length: %i\n",l->length);
}
