/* Copyright (c) 2005, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// mazegen.c header file
#ifndef MAZEGEN_H
#define MAZEGEN_H



//// BITVECTORS ////
// Set these in the individual elements of **path in a MAZE
////////

// Room Connectors
#define N 1
#define E 2
#define S 4
#define W 8

// Reachability (set to R, reachable, when hollowed and reached by digger)
#define R 16

// Is part of the solution (changes the way it's rendered on-screen)
#define SOL 32

#define START 64

//// STRUCTS ////
// Contain all sorts of blocks of data
////////

// Vectors, useful to have when working in non-1D space
typedef struct {
  int x;
  int y;
} vec;

// The maze itself.  **path is a 2D array that has bitvectors to represent each cell in the maze, with values as defined above.
typedef struct {
  vec enter;
  vec exit;
  unsigned int x;
  unsigned int y;
  unsigned int cflags; // Carver flags, to be used as the digger deems fit.
  char **path; // N and E are ignored in this flag.  Instead, set S and W in the opposite cell.
} MAZE;

//// LINKED LIST ////
// Used for lots of algorithms, for example to keep track of which cells can still be dug from.
////////

struct CELL {
  int x;
  int y;
  int value;
  struct CELL *next;
};

typedef struct {
  int length;
  struct CELL *cell;
} LLIST;


LLIST *newllist();
void destroyllist(LLIST *l);
void addlink(LLIST *l, int pos, int x, int y, int value);
void remlink(LLIST *l, int pos); // Passing a negative pos to add/remlink will cause it to count backwards from the end of the list instead.
struct CELL *getlink(LLIST *l, int pos);
void printllist(LLIST *l);

void dispmaze(MAZE *m, int fx, int fy, char *header);
MAZE *newmaze(int x, int y);

char *findvaliddirs(MAZE *m, int cx, int cy, int r);



vec dirtovec(int dir);

char edge(MAZE *m, int x, int y, int dir);

int mget(MAZE *m, int x, int y);
void mset(MAZE *m, int x, int y, int val);
void munset(MAZE *m, int x, int y, int val);

void dig(MAZE *m, int x, int y, int dir);


void D(char *str);


// Include the header files for any used algorithms
#include "mazegen_backtracer.h"
#include "mazegen_growtree.h"

#endif
