/* Copyright (c) 2005, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "mazegen.h"


MAZE *mazegen_backtracer(int x, int y, vec start, vec end, int step) {
  // Start by creating a maze to live in.
  MAZE *m = newmaze(x,y);
  m->enter.x = start.x;
  m->enter.y = start.y;
  m->exit.x = end.x;
  m->exit.y = end.y;

  // Start at the entrance, head to exit.
  int cx, cy;
  cx = m->enter.x;
  cy = m->enter.y;
  mset(m,cx,cy,R);
  int exitfound  = 0;
  
  char *stack; // Keep track of all our movements
  stack = malloc(sizeof(char)*x*y+5);
  int spos = 0; // Position on the stack.

  int steps = 0;
  
  int bt = 0; // Backtracing?  So that we don't pop holes down everything.

  char lastdir = -1;


  while (1) {
    // Find all valid directions (undug, not edge of earth) from here to move
    // (now offloaded to an outside script)
    char *dirs = findvaliddirs(m,cx,cy,0);
    int p = dirs[0];
    dirs+=1; // Point to the string itself, not its length+string
    
    if (p == 0 && cx == m->enter.x && cy == m->enter.y) {
      return m; // At the beginning, nowhere to go.  Bail.
    }

    if (p > 0 && !(cx == m->exit.x && cy == m->exit.y) ) { // Got some adjacent cells, let's move to a random one.
      
      bt = 0; // definately not backtracing now.
      int dir = -1;
      if (0 && rand()%2) { // Half chance of repeating last step, if possible.
	int c;
	dir = -1;
	for (c = 0; c < p; c++) {
	  if (dirs[c] == lastdir) {
	    dir = lastdir;
	  }
	}
	if (dir == -1) {
	  dir = dirs[rand()%p];
	}
      } else {
	dir = dirs[rand()%p];
      }
      dig(m, cx,cy, dir);
      lastdir = dir;
      steps++;
      stack[spos++] = dir;
      vec d = dirtovec(dir);
      cx += d.x;
      cy += d.y;
      //      if (!exitfound)
      //	mset(m,cx,cy,SOL);
      if (cx == m->exit.x && cy == m->exit.y)
	exitfound = 1;
      
      if (step) {
	char *buf;
	buf = malloc(80);
	snprintf(buf,80,"Generating a %ix%i maze... (%03i steps taken, stacklen %03i)",x,y,steps,spos);
	dispmaze(m, cx,cy, buf);
	free(buf);
	//usleep(33333);
	usleep(22222);
      }
    } // if (p > 0)
    else { // No adjecent cells, backtrace instead.


      if (!bt) { // If we're not already backtracing, let's pop a hole.
	bt = 1;
	if (rand()%10 == 5) {
	  if ( !( cx == m->exit.x && cy == m->exit.y ) ) { // Only one exitpath
	    // dig a random way, to create some loops.
	    char *dirs = findvaliddirs(m,cx,cy,1);
	    int p = dirs[0];
	    dirs+=1;
	    dig(m, cx,cy, dirs[rand()%p]);
	  }
	}
      }
      
      // Now, to backtrace.

      int dir = stack[--spos];

      // Not part of the solution.
      //      if (!exitfound)
	//	munset(m,cx,cy,SOL);

      //      spos--;
      vec d = dirtovec(dir);
      cx -= d.x;
      cy -= d.y;
      
      if (step) {
	char *buf;
	buf = malloc(80);
	snprintf(buf,80,"Backtracing... stacklen %03i",spos);
	dispmaze(m, cx,cy, buf);
	free(buf);
	usleep(11111);
      }
    }
      
    
  }
  return m;
}

