/* Copyright (c) 2005, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "mazegen.h"

#define LOG(x) fprintf(log, x)

MAZE *mazegen_growtree(int x, int y, vec start, vec end, int step) {
  FILE *log = fopen("DEBUG.txt", "w");
  // Start by creating a maze to live in.
  MAZE *m = newmaze(x,y);
  m->enter.x = start.x;
  m->enter.y = start.y;
  m->exit.x = end.x;
  m->exit.y = end.y;
  
  // Start at the entrance, head to exit.
  mset(m,m->enter.x,m->enter.y,R|START);
  
  int steps = 0;

  LLIST *ll = newllist();
  // Our linked list of valid places to dig from.  Let's put the starting position in it first, so we have something to work with.
  addlink(ll,0,m->enter.x, m->enter.y, 0);

  int exitfound = 0;

  while (1) {

    // Have we any cells to dig in?
    if (ll->length == 0) {
      destroyllist(ll);
      return m; // Nope, send back the completed maze
    }

    // Pick a random cell to dig from
    int rnum;
    int max;
    if (exitfound) {
      max = ll->length / 2;
    } else {
      max = ll->length / 40;
    }
    if (max < 1)
      max = 1;
    rnum = rand()%max; // Well, haven't found the exit yet, let's just meander aimlessly until we do, for a good random solution.

    vec pos;
    int flip = 1;
    if (exitfound && rand()%2) // Randomly start pulling from the opposite end of the stack after we've found the exit
	flip = -1;
    rnum *= flip;

    struct CELL *hi = getlink(ll, rnum);
    pos.x = hi->x;
    pos.y = hi->y;
    

    // Find all valid directions (undug, not edge of earth) from here to move
    // (now offloaded to an outside script)
    char *dirs = findvaliddirs(m,pos.x,pos.y,0);

    int p = dirs[0];
    dirs+=1; // Point to the string itself, not its length+string

    if (p <= 1) { // That was the only move we could make, that cell shouldn't be considered in the future
      remlink(ll,rnum);
    }

    
    if (p > 0 && !(pos.x == m->exit.x && pos.y == m->exit.y) ) { // Got some adjacent cells, let's move to a random one.
      int dir;
      
      if (rand()%5 == 2) { // Bias toward the exit sometimes
	LOG("Findign a biased direction to move\n");
	char *bias; int l = 0;
	bias = malloc(8);
	int x;
	if (pos.x < m->exit.x) {
	  LOG("Maybe east...\n");
	  for (x = 0; x < p; x++) {
	    if (dirs[x] == E) {
	      LOG("East works\n");
	      bias[l] = E; l++;
	    }
	  }
	}
	if (pos.x > m->exit.x) {
	  LOG("Maybe west...\n");
	  for (x = 0; x < p; x++) {
	    if (dirs[x] == W) {
	      LOG("west works\n");
	      bias[l] = W; l++;
	    }
	  }
	}
	if (pos.y > m->exit.y) {
	  LOG("Maybe north\n");
	  for (x = 0; x < p; x++) {
	    if (dirs[x] == N) {
	      LOG("n works\n");
	      bias[l] = N; l++;
	    }
	  }
	}
	if (pos.y < m->exit.y) {	
	  LOG("maybe S\n");
	  for (x = 0; x < p; x++) {
	    if (dirs[x] == S) {
	      LOG("sure\n");
	      bias[l] = S; l++;
	    }
	  }
	}
	if (l != 0) {
	  dir = bias[rand()%l];
	} else {
	  dir = dirs[rand()%p];
	}
	free(bias);
      } else {
	dir = dirs[rand()%p];
      }
      free(--dirs);
	
      dig(m, pos.x,pos.y, dir);

      steps++;

      

      vec d = dirtovec(dir);
      pos.x += d.x;
      pos.y += d.y;
      
      // Add this new dig to our linked list if it's not an exit (exit should never dig out, only be dug into once)
      if ( ! (pos.x == m->exit.x && pos.y == m->exit.y) ) {
	addlink(ll,0,pos.x,pos.y,0);
      } else {
	exitfound = 1; // Found the exit, let's just start makign this maze pretty crazy now.
      }
      
      if (step) {
	char *buf;
	buf = malloc(80);
	snprintf(buf,80,"Generating a %ix%i maze... (%03i steps taken, %03i stack size)",x,y,steps, ll->length);
	dispmaze(m, pos.x,pos.y, buf);
	free(buf);
	usleep(22222);
      }
      //      usleep(44444);
    } // if (p > 0)
    //    else if (p == 0) { // Nowhere to go here, remove that entry from our array.
      // This shouldn't even be possbile anymore, since we check earlier when finding valid paths to move.
    //      remlink(ll,rnum);
    //    }
    
    
  }
  destroyllist(ll);
  return m;
}

